package org.apache.bcel.generic;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.Hashtable;
import java.util.List;
import java.util.Objects;
import java.util.Stack;
import org.apache.bcel.classfile.AnnotationEntry;
import org.apache.bcel.classfile.Annotations;
import org.apache.bcel.classfile.Attribute;
import org.apache.bcel.classfile.Code;
import org.apache.bcel.classfile.CodeException;
import org.apache.bcel.classfile.ExceptionTable;
import org.apache.bcel.classfile.LineNumber;
import org.apache.bcel.classfile.LineNumberTable;
import org.apache.bcel.classfile.LocalVariable;
import org.apache.bcel.classfile.LocalVariableTable;
import org.apache.bcel.classfile.LocalVariableTypeTable;
import org.apache.bcel.classfile.Method;
import org.apache.bcel.classfile.ParameterAnnotationEntry;
import org.apache.bcel.classfile.ParameterAnnotations;
import org.apache.bcel.classfile.RuntimeVisibleParameterAnnotations;
import org.apache.bcel.classfile.Utility;
import org.apache.bcel.util.BCELComparator;
import org.apache.commons.lang3.ArrayUtils;

/* loaded from: org_apache_bcel_generic.jar:org/apache/bcel/generic/MethodGen.class */
public class MethodGen extends FieldGenOrMethodGen {
    private static BCELComparator bcelComparator = new BCELComparator() { // from class: org.apache.bcel.generic.MethodGen.1
        public boolean equals(Object o1, Object o2) {
            FieldGenOrMethodGen THIS = (FieldGenOrMethodGen) o1;
            FieldGenOrMethodGen THAT = (FieldGenOrMethodGen) o2;
            return Objects.equals(THIS.getName(), THAT.getName()) && Objects.equals(THIS.getSignature(), THAT.getSignature());
        }

        public int hashCode(Object o) {
            FieldGenOrMethodGen THIS = (FieldGenOrMethodGen) o;
            return THIS.getSignature().hashCode() ^ THIS.getName().hashCode();
        }
    };
    private String className;
    private Type[] argTypes;
    private String[] argNames;
    private int maxLocals;
    private int maxStack;
    private InstructionList il;
    private boolean stripAttributes;
    private LocalVariableTypeTable localVariableTypeTable;
    private final List<LocalVariableGen> variableList;
    private final List<LineNumberGen> lineNumberList;
    private final List<CodeExceptionGen> exceptionList;
    private final List<String> throwsList;
    private final List<Attribute> codeAttrsList;
    private List<AnnotationEntryGen>[] paramAnnotations;
    private boolean hasParameterAnnotations;
    private boolean haveUnpackedParameterAnnotations;
    private List<MethodObserver> observers;

    /* JADX INFO: Access modifiers changed from: package-private */
    /* loaded from: org_apache_bcel_generic.jar:org/apache/bcel/generic/MethodGen$BranchStack.class */
    public static final class BranchStack {
        private final Stack<BranchTarget> branchTargets = new Stack<>();
        private final Hashtable<InstructionHandle, BranchTarget> visitedTargets = new Hashtable<>();

        BranchStack() {
        }

        public BranchTarget pop() {
            if (!this.branchTargets.empty()) {
                return this.branchTargets.pop();
            }
            return null;
        }

        public void push(InstructionHandle target, int stackDepth) {
            if (visited(target)) {
                return;
            }
            this.branchTargets.push(visit(target, stackDepth));
        }

        private BranchTarget visit(InstructionHandle target, int stackDepth) {
            BranchTarget bt = new BranchTarget(target, stackDepth);
            this.visitedTargets.put(target, bt);
            return bt;
        }

        private boolean visited(InstructionHandle target) {
            return this.visitedTargets.get(target) != null;
        }
    }

    /* JADX INFO: Access modifiers changed from: package-private */
    /* loaded from: org_apache_bcel_generic.jar:org/apache/bcel/generic/MethodGen$BranchTarget.class */
    public static final class BranchTarget {
        final InstructionHandle target;
        final int stackDepth;

        BranchTarget(InstructionHandle target, int stackDepth) {
            this.target = target;
            this.stackDepth = stackDepth;
        }
    }

    private static byte[] getByteCodes(Method method) {
        Code code = method.getCode();
        if (code == null) {
            throw new IllegalStateException(String.format("The method '%s' has no code.", method));
        }
        return code.getCode();
    }

    public static BCELComparator getComparator() {
        return bcelComparator;
    }

    public static int getMaxStack(ConstantPoolGen cp, InstructionList il, CodeExceptionGen[] et) {
        BranchTarget bt;
        BranchStack branchTargets = new BranchStack();
        for (CodeExceptionGen element : et) {
            InstructionHandle handlerPc = element.getHandlerPC();
            if (handlerPc != null) {
                branchTargets.push(handlerPc, 1);
            }
        }
        int stackDepth = 0;
        int maxStackDepth = 0;
        InstructionHandle ih = il.getStart();
        while (ih != null) {
            Instruction instruction = ih.getInstruction();
            short opcode = instruction.getOpcode();
            int delta = instruction.produceStack(cp) - instruction.consumeStack(cp);
            stackDepth += delta;
            if (stackDepth > maxStackDepth) {
                maxStackDepth = stackDepth;
            }
            if (instruction instanceof BranchInstruction) {
                BranchInstruction branch = (BranchInstruction) instruction;
                if (instruction instanceof Select) {
                    Select select = (Select) branch;
                    InstructionHandle[] targets = select.getTargets();
                    for (InstructionHandle target : targets) {
                        branchTargets.push(target, stackDepth);
                    }
                    ih = null;
                } else if (!(branch instanceof IfInstruction)) {
                    if (opcode == 168 || opcode == 201) {
                        branchTargets.push(ih.getNext(), stackDepth - 1);
                    }
                    ih = null;
                }
                branchTargets.push(branch.getTarget(), stackDepth);
            } else if (opcode == 191 || opcode == 169 || (opcode >= 172 && opcode <= 177)) {
                ih = null;
            }
            if (ih != null) {
                ih = ih.getNext();
            }
            if (ih == null && (bt = branchTargets.pop()) != null) {
                ih = bt.target;
                stackDepth = bt.stackDepth;
            }
        }
        return maxStackDepth;
    }

    public static void setComparator(BCELComparator comparator) {
        bcelComparator = comparator;
    }

    public MethodGen(int accessFlags, Type returnType, Type[] argTypes, String[] argNames, String methodName, String className, InstructionList il, ConstantPoolGen cp) {
        super(accessFlags);
        this.variableList = new ArrayList();
        this.lineNumberList = new ArrayList();
        this.exceptionList = new ArrayList();
        this.throwsList = new ArrayList();
        this.codeAttrsList = new ArrayList();
        setType(returnType);
        setArgumentTypes(argTypes);
        setArgumentNames(argNames);
        setName(methodName);
        setClassName(className);
        setInstructionList(il);
        setConstantPool(cp);
        boolean abstract_ = isAbstract() || isNative();
        InstructionHandle start = null;
        if (!abstract_) {
            start = il.getStart();
            if (!isStatic() && className != null) {
                addLocalVariable("this", ObjectType.getInstance(className), start, null);
            }
        }
        if (argTypes != null) {
            int size = argTypes.length;
            for (Type argType : argTypes) {
                if (Type.VOID == argType) {
                    throw new ClassGenException("'void' is an illegal argument type for a method");
                }
            }
            if (argNames != null) {
                if (size != argNames.length) {
                    throw new ClassGenException("Mismatch in argument array lengths: " + size + " vs. " + argNames.length);
                }
            } else {
                argNames = new String[size];
                for (int i = 0; i < size; i++) {
                    argNames[i] = "arg" + i;
                }
                setArgumentNames(argNames);
            }
            if (!abstract_) {
                for (int i2 = 0; i2 < size; i2++) {
                    addLocalVariable(argNames[i2], argTypes[i2], start, null);
                }
            }
        }
    }

    public MethodGen(Method method, String className, ConstantPoolGen cp) {
        this(method.getAccessFlags(), Type.getReturnType(method.getSignature()), Type.getArgumentTypes(method.getSignature()), null, method.getName(), className, (method.getAccessFlags() & 1280) == 0 ? new InstructionList(getByteCodes(method)) : null, cp);
        InstructionHandle prev;
        for (Code code : method.getAttributes()) {
            if (!(code instanceof Code)) {
                if (code instanceof ExceptionTable) {
                    Collections.addAll(this.throwsList, ((ExceptionTable) code).getExceptionNames());
                } else if (code instanceof Annotations) {
                    Annotations runtimeAnnotations = (Annotations) code;
                    runtimeAnnotations.forEach(element -> {
                        addAnnotationEntry(new AnnotationEntryGen(element, cp, false));
                    });
                } else {
                    addAttribute(code);
                }
            } else {
                Code c = code;
                setMaxStack(c.getMaxStack());
                setMaxLocals(c.getMaxLocals());
                CodeException[] ces = c.getExceptionTable();
                if (ces != null) {
                    for (CodeException ce : ces) {
                        int type = ce.getCatchType();
                        ObjectType cType = null;
                        if (type > 0) {
                            String cen = method.getConstantPool().getConstantString(type, (byte) 7);
                            cType = ObjectType.getInstance(cen);
                        }
                        int endPc = ce.getEndPC();
                        int length = getByteCodes(method).length;
                        if (length == endPc) {
                            prev = this.il.getEnd();
                        } else {
                            InstructionHandle end = this.il.findHandle(endPc);
                            prev = end.getPrev();
                        }
                        InstructionHandle end2 = prev;
                        addExceptionHandler(this.il.findHandle(ce.getStartPC()), end2, this.il.findHandle(ce.getHandlerPC()), cType);
                    }
                }
                for (LineNumberTable lineNumberTable : c.getAttributes()) {
                    if (lineNumberTable instanceof LineNumberTable) {
                        lineNumberTable.forEach(l -> {
                            InstructionHandle ih = this.il.findHandle(l.getStartPC());
                            if (ih != null) {
                                addLineNumber(ih, l.getLineNumber());
                            }
                        });
                    } else if (lineNumberTable instanceof LocalVariableTable) {
                        updateLocalVariableTable((LocalVariableTable) lineNumberTable);
                    } else if (lineNumberTable instanceof LocalVariableTypeTable) {
                        this.localVariableTypeTable = lineNumberTable.copy(cp.getConstantPool());
                    } else {
                        addCodeAttribute(lineNumberTable);
                    }
                }
            }
        }
    }

    public void addAnnotationsAsAttribute(ConstantPoolGen cp) {
        addAll(AnnotationEntryGen.getAnnotationAttributes(cp, super.getAnnotationEntries()));
    }

    public void addCodeAttribute(Attribute a) {
        this.codeAttrsList.add(a);
    }

    public void addException(String className) {
        this.throwsList.add(className);
    }

    public CodeExceptionGen addExceptionHandler(InstructionHandle startPc, InstructionHandle endPc, InstructionHandle handlerPc, ObjectType catchType) {
        if (startPc == null || endPc == null || handlerPc == null) {
            throw new ClassGenException("Exception handler target is null instruction");
        }
        CodeExceptionGen c = new CodeExceptionGen(startPc, endPc, handlerPc, catchType);
        this.exceptionList.add(c);
        return c;
    }

    public LineNumberGen addLineNumber(InstructionHandle ih, int srcLine) {
        LineNumberGen l = new LineNumberGen(ih, srcLine);
        this.lineNumberList.add(l);
        return l;
    }

    public LocalVariableGen addLocalVariable(String name, Type type, InstructionHandle start, InstructionHandle end) {
        return addLocalVariable(name, type, this.maxLocals, start, end);
    }

    public LocalVariableGen addLocalVariable(String name, Type type, int slot, InstructionHandle start, InstructionHandle end) {
        return addLocalVariable(name, type, slot, start, end, slot);
    }

    public LocalVariableGen addLocalVariable(String name, Type type, int slot, InstructionHandle start, InstructionHandle end, int origIndex) {
        byte t = type.getType();
        if (t != 16) {
            int add = type.getSize();
            if (slot + add > this.maxLocals) {
                this.maxLocals = slot + add;
            }
            LocalVariableGen l = new LocalVariableGen(slot, name, type, start, end, origIndex);
            int i = this.variableList.indexOf(l);
            if (i >= 0) {
                this.variableList.set(i, l);
            } else {
                this.variableList.add(l);
            }
            return l;
        }
        throw new IllegalArgumentException("Can not use " + type + " as type for local variable");
    }

    public void addObserver(MethodObserver o) {
        if (this.observers == null) {
            this.observers = new ArrayList();
        }
        this.observers.add(o);
    }

    public void addParameterAnnotation(int parameterIndex, AnnotationEntryGen annotation) {
        ensureExistingParameterAnnotationsUnpacked();
        if (!this.hasParameterAnnotations) {
            List<AnnotationEntryGen>[] parmList = new List[this.argTypes.length];
            this.paramAnnotations = parmList;
            this.hasParameterAnnotations = true;
        }
        List<AnnotationEntryGen> existingAnnotations = this.paramAnnotations[parameterIndex];
        if (existingAnnotations != null) {
            existingAnnotations.add(annotation);
            return;
        }
        List<AnnotationEntryGen> l = new ArrayList<>();
        l.add(annotation);
        this.paramAnnotations[parameterIndex] = l;
    }

    public void addParameterAnnotationsAsAttribute(ConstantPoolGen cp) {
        Attribute[] attrs;
        if (this.hasParameterAnnotations && (attrs = AnnotationEntryGen.getParameterAnnotationAttributes(cp, this.paramAnnotations)) != null) {
            addAll(attrs);
        }
    }

    private Attribute[] addRuntimeAnnotationsAsAttribute(ConstantPoolGen cp) {
        Attribute[] attrs = AnnotationEntryGen.getAnnotationAttributes(cp, super.getAnnotationEntries());
        addAll(attrs);
        return attrs;
    }

    private Attribute[] addRuntimeParameterAnnotationsAsAttribute(ConstantPoolGen cp) {
        if (!this.hasParameterAnnotations) {
            return Attribute.EMPTY_ARRAY;
        }
        Attribute[] attrs = AnnotationEntryGen.getParameterAnnotationAttributes(cp, this.paramAnnotations);
        addAll(attrs);
        return attrs;
    }

    private void adjustLocalVariableTypeTable(LocalVariableTable lvt) {
        LocalVariable[] lv = lvt.getLocalVariableTable();
        for (LocalVariable element : this.localVariableTypeTable.getLocalVariableTypeTable()) {
            int length = lv.length;
            int i = 0;
            while (true) {
                if (i < length) {
                    LocalVariable l = lv[i];
                    if (!element.getName().equals(l.getName()) || element.getIndex() != l.getOrigIndex()) {
                        i++;
                    } else {
                        element.setLength(l.getLength());
                        element.setStartPC(l.getStartPC());
                        element.setIndex(l.getIndex());
                        break;
                    }
                }
            }
        }
    }

    public MethodGen copy(String className, ConstantPoolGen cp) {
        Method m = ((MethodGen) clone()).getMethod();
        MethodGen mg = new MethodGen(m, className, super.getConstantPool());
        if (super.getConstantPool() != cp) {
            mg.setConstantPool(cp);
            mg.getInstructionList().replaceConstantPool(super.getConstantPool(), cp);
        }
        return mg;
    }

    private void ensureExistingParameterAnnotationsUnpacked() {
        if (this.haveUnpackedParameterAnnotations) {
            return;
        }
        ParameterAnnotations paramAnnVisAttr = null;
        ParameterAnnotations paramAnnInvisAttr = null;
        for (ParameterAnnotations parameterAnnotations : getAttributes()) {
            if (parameterAnnotations instanceof ParameterAnnotations) {
                if (!this.hasParameterAnnotations) {
                    List<AnnotationEntryGen>[] parmList = new List[this.argTypes.length];
                    this.paramAnnotations = parmList;
                    Arrays.setAll(this.paramAnnotations, i -> {
                        return new ArrayList();
                    });
                }
                this.hasParameterAnnotations = true;
                ParameterAnnotations rpa = parameterAnnotations;
                if (rpa instanceof RuntimeVisibleParameterAnnotations) {
                    paramAnnVisAttr = rpa;
                } else {
                    paramAnnInvisAttr = rpa;
                }
                ParameterAnnotationEntry[] parameterAnnotationEntries = rpa.getParameterAnnotationEntries();
                for (int j = 0; j < parameterAnnotationEntries.length; j++) {
                    ParameterAnnotationEntry immutableArray = rpa.getParameterAnnotationEntries()[j];
                    List<AnnotationEntryGen> mutable = makeMutableVersion(immutableArray.getAnnotationEntries());
                    this.paramAnnotations[j].addAll(mutable);
                }
            }
        }
        if (paramAnnVisAttr != null) {
            removeAttribute(paramAnnVisAttr);
        }
        if (paramAnnInvisAttr != null) {
            removeAttribute(paramAnnInvisAttr);
        }
        this.haveUnpackedParameterAnnotations = true;
    }

    public boolean equals(Object obj) {
        return bcelComparator.equals(this, obj);
    }

    public List<AnnotationEntryGen> getAnnotationsOnParameter(int i) {
        ensureExistingParameterAnnotationsUnpacked();
        if (!this.hasParameterAnnotations || i > this.argTypes.length) {
            return null;
        }
        return this.paramAnnotations[i];
    }

    public String getArgumentName(int i) {
        return this.argNames[i];
    }

    public String[] getArgumentNames() {
        return (String[]) this.argNames.clone();
    }

    public Type getArgumentType(int i) {
        return this.argTypes[i];
    }

    public Type[] getArgumentTypes() {
        return (Type[]) this.argTypes.clone();
    }

    public String getClassName() {
        return this.className;
    }

    public Attribute[] getCodeAttributes() {
        return (Attribute[]) this.codeAttrsList.toArray(Attribute.EMPTY_ARRAY);
    }

    private CodeException[] getCodeExceptions() {
        int size = this.exceptionList.size();
        CodeException[] cExc = new CodeException[size];
        Arrays.setAll(cExc, i -> {
            return this.exceptionList.get(i).getCodeException(super.getConstantPool());
        });
        return cExc;
    }

    public CodeExceptionGen[] getExceptionHandlers() {
        return (CodeExceptionGen[]) this.exceptionList.toArray(CodeExceptionGen.EMPTY_ARRAY);
    }

    public String[] getExceptions() {
        return (String[]) this.throwsList.toArray(ArrayUtils.EMPTY_STRING_ARRAY);
    }

    private ExceptionTable getExceptionTable(ConstantPoolGen cp) {
        int size = this.throwsList.size();
        int[] ex = new int[size];
        Arrays.setAll(ex, i -> {
            return cp.addClass(this.throwsList.get(i));
        });
        return new ExceptionTable(cp.addUtf8("Exceptions"), 2 + (2 * size), ex, cp.getConstantPool());
    }

    public InstructionList getInstructionList() {
        return this.il;
    }

    public LineNumberGen[] getLineNumbers() {
        return (LineNumberGen[]) this.lineNumberList.toArray(LineNumberGen.EMPTY_ARRAY);
    }

    public LineNumberTable getLineNumberTable(ConstantPoolGen cp) {
        int size = this.lineNumberList.size();
        LineNumber[] ln = new LineNumber[size];
        Arrays.setAll(ln, i -> {
            return this.lineNumberList.get(i).getLineNumber();
        });
        return new LineNumberTable(cp.addUtf8("LineNumberTable"), 2 + (ln.length * 4), ln, cp.getConstantPool());
    }

    public LocalVariableGen[] getLocalVariables() {
        int size = this.variableList.size();
        LocalVariableGen[] lg = new LocalVariableGen[size];
        this.variableList.toArray(lg);
        for (int i = 0; i < size; i++) {
            if (lg[i].getStart() == null && this.il != null) {
                lg[i].setStart(this.il.getStart());
            }
            if (lg[i].getEnd() == null && this.il != null) {
                lg[i].setEnd(this.il.getEnd());
            }
        }
        if (size > 1) {
            Arrays.sort(lg, Comparator.comparingInt((v0) -> {
                return v0.getIndex();
            }));
        }
        return lg;
    }

    public LocalVariableTable getLocalVariableTable(ConstantPoolGen cp) {
        LocalVariableGen[] lg = getLocalVariables();
        int size = lg.length;
        LocalVariable[] lv = new LocalVariable[size];
        Arrays.setAll(lv, i -> {
            return lg[i].getLocalVariable(cp);
        });
        return new LocalVariableTable(cp.addUtf8("LocalVariableTable"), 2 + (lv.length * 10), lv, cp.getConstantPool());
    }

    public LocalVariableTypeTable getLocalVariableTypeTable() {
        return this.localVariableTypeTable;
    }

    public int getMaxLocals() {
        return this.maxLocals;
    }

    public int getMaxStack() {
        return this.maxStack;
    }

    public Method getMethod() {
        String signature = getSignature();
        ConstantPoolGen cp = super.getConstantPool();
        int nameIndex = cp.addUtf8(super.getName());
        int signatureIndex = cp.addUtf8(signature);
        byte[] byteCode = this.il != null ? this.il.getByteCode() : null;
        LineNumberTable lnt = null;
        LocalVariableTable lvt = null;
        if (!this.variableList.isEmpty() && !this.stripAttributes) {
            updateLocalVariableTable(getLocalVariableTable(cp));
            LocalVariableTable localVariableTable = getLocalVariableTable(cp);
            lvt = localVariableTable;
            addCodeAttribute(localVariableTable);
        }
        if (this.localVariableTypeTable != null) {
            if (lvt != null) {
                adjustLocalVariableTypeTable(lvt);
            }
            addCodeAttribute(this.localVariableTypeTable);
        }
        if (!this.lineNumberList.isEmpty() && !this.stripAttributes) {
            LineNumberTable lineNumberTable = getLineNumberTable(cp);
            lnt = lineNumberTable;
            addCodeAttribute(lineNumberTable);
        }
        Attribute[] codeAttrs = getCodeAttributes();
        int attrsLen = 0;
        for (Attribute codeAttr : codeAttrs) {
            attrsLen += codeAttr.getLength() + 6;
        }
        CodeException[] cExc = getCodeExceptions();
        int excLen = cExc.length * 8;
        Attribute attribute = null;
        if (byteCode != null && !isAbstract() && !isNative()) {
            Attribute[] attributes = getAttributes();
            for (Attribute a : attributes) {
                if (a instanceof Code) {
                    removeAttribute(a);
                }
            }
            attribute = new Code(cp.addUtf8("Code"), 8 + byteCode.length + 2 + excLen + 2 + attrsLen, this.maxStack, this.maxLocals, byteCode, cExc, codeAttrs, cp.getConstantPool());
            addAttribute(attribute);
        }
        Attribute[] annotations = addRuntimeAnnotationsAsAttribute(cp);
        Attribute[] parameterAnnotations = addRuntimeParameterAnnotationsAsAttribute(cp);
        ExceptionTable et = null;
        if (!this.throwsList.isEmpty()) {
            ExceptionTable exceptionTable = getExceptionTable(cp);
            et = exceptionTable;
            addAttribute(exceptionTable);
        }
        Method m = new Method(super.getAccessFlags(), nameIndex, signatureIndex, getAttributes(), cp.getConstantPool());
        if (lvt != null) {
            removeCodeAttribute(lvt);
        }
        if (this.localVariableTypeTable != null) {
            removeCodeAttribute(this.localVariableTypeTable);
        }
        if (lnt != null) {
            removeCodeAttribute(lnt);
        }
        if (attribute != null) {
            removeAttribute(attribute);
        }
        if (et != null) {
            removeAttribute(et);
        }
        removeRuntimeAttributes(annotations);
        removeRuntimeAttributes(parameterAnnotations);
        return m;
    }

    public Type getReturnType() {
        return getType();
    }

    @Override // org.apache.bcel.generic.FieldGenOrMethodGen
    public String getSignature() {
        return Type.getMethodSignature(super.getType(), this.argTypes);
    }

    public int hashCode() {
        return bcelComparator.hashCode(this);
    }

    private List<AnnotationEntryGen> makeMutableVersion(AnnotationEntry[] mutableArray) {
        List<AnnotationEntryGen> result = new ArrayList<>();
        for (AnnotationEntry element : mutableArray) {
            result.add(new AnnotationEntryGen(element, getConstantPool(), false));
        }
        return result;
    }

    public void removeCodeAttribute(Attribute a) {
        this.codeAttrsList.remove(a);
    }

    public void removeCodeAttributes() {
        this.localVariableTypeTable = null;
        this.codeAttrsList.clear();
    }

    public void removeException(String c) {
        this.throwsList.remove(c);
    }

    public void removeExceptionHandler(CodeExceptionGen c) {
        this.exceptionList.remove(c);
    }

    public void removeExceptionHandlers() {
        this.exceptionList.clear();
    }

    public void removeExceptions() {
        this.throwsList.clear();
    }

    public void removeLineNumber(LineNumberGen l) {
        this.lineNumberList.remove(l);
    }

    public void removeLineNumbers() {
        this.lineNumberList.clear();
    }

    public void removeLocalVariable(LocalVariableGen l) {
        l.dispose();
        this.variableList.remove(l);
    }

    public void removeLocalVariables() {
        this.variableList.forEach((v0) -> {
            v0.dispose();
        });
        this.variableList.clear();
    }

    public void removeLocalVariableTypeTable() {
        this.localVariableTypeTable = null;
    }

    public void removeNOPs() {
        if (this.il != null) {
            InstructionHandle start = this.il.getStart();
            while (true) {
                InstructionHandle ih = start;
                if (ih != null) {
                    InstructionHandle next = ih.getNext();
                    if (next != null && (ih.getInstruction() instanceof NOP)) {
                        try {
                            this.il.delete(ih);
                        } catch (TargetLostException e) {
                            for (InstructionHandle target : e.getTargets()) {
                                for (InstructionTargeter targeter : target.getTargeters()) {
                                    targeter.updateTarget(target, next);
                                }
                            }
                        }
                    }
                    start = next;
                } else {
                    return;
                }
            }
        }
    }

    public void removeObserver(MethodObserver o) {
        if (this.observers != null) {
            this.observers.remove(o);
        }
    }

    public void removeRuntimeAttributes(Attribute[] attrs) {
        for (Attribute attr : attrs) {
            removeAttribute(attr);
        }
    }

    public void setArgumentName(int i, String name) {
        this.argNames[i] = name;
    }

    public void setArgumentNames(String[] argNames) {
        this.argNames = argNames;
    }

    public void setArgumentType(int i, Type type) {
        this.argTypes[i] = type;
    }

    public void setArgumentTypes(Type[] argTypes) {
        this.argTypes = argTypes;
    }

    public void setClassName(String className) {
        this.className = className;
    }

    public void setInstructionList(InstructionList il) {
        this.il = il;
    }

    public void setMaxLocals() {
        int index;
        if (this.il != null) {
            int max = isStatic() ? 0 : 1;
            if (this.argTypes != null) {
                for (Type argType : this.argTypes) {
                    max += argType.getSize();
                }
            }
            InstructionHandle start = this.il.getStart();
            while (true) {
                InstructionHandle ih = start;
                if (ih != null) {
                    Cloneable instruction = ih.getInstruction();
                    if (((instruction instanceof LocalVariableInstruction) || (instruction instanceof RET) || (instruction instanceof IINC)) && (index = ((IndexedInstruction) instruction).getIndex() + ((TypedInstruction) instruction).getType(super.getConstantPool()).getSize()) > max) {
                        max = index;
                    }
                    start = ih.getNext();
                } else {
                    this.maxLocals = max;
                    return;
                }
            }
        } else {
            this.maxLocals = 0;
        }
    }

    public void setMaxLocals(int m) {
        this.maxLocals = m;
    }

    public void setMaxStack() {
        if (this.il != null) {
            this.maxStack = getMaxStack(super.getConstantPool(), this.il, getExceptionHandlers());
        } else {
            this.maxStack = 0;
        }
    }

    public void setMaxStack(int m) {
        this.maxStack = m;
    }

    public void setReturnType(Type returnType) {
        setType(returnType);
    }

    public void stripAttributes(boolean flag) {
        this.stripAttributes = flag;
    }

    public final String toString() {
        String access = Utility.accessToString(super.getAccessFlags());
        String signature = Type.getMethodSignature(super.getType(), this.argTypes);
        StringBuilder buf = new StringBuilder(Utility.methodSignatureToString(signature, super.getName(), access, true, getLocalVariableTable(super.getConstantPool())));
        for (Attribute a : getAttributes()) {
            if (!(a instanceof Code) && !(a instanceof ExceptionTable)) {
                buf.append(" [").append(a).append("]");
            }
        }
        if (!this.throwsList.isEmpty()) {
            for (String throwsDescriptor : this.throwsList) {
                buf.append("\n\t\tthrows ").append(throwsDescriptor);
            }
        }
        return buf.toString();
    }

    public void update() {
        if (this.observers != null) {
            for (MethodObserver observer : this.observers) {
                observer.notify(this);
            }
        }
    }

    private void updateLocalVariableTable(LocalVariableTable a) {
        removeLocalVariables();
        for (LocalVariable l : a.getLocalVariableTable()) {
            InstructionHandle start = this.il.findHandle(l.getStartPC());
            InstructionHandle end = this.il.findHandle(l.getStartPC() + l.getLength());
            if (null == start) {
                start = this.il.getStart();
            }
            addLocalVariable(l.getName(), Type.getType(l.getSignature()), l.getIndex(), start, end, l.getOrigIndex());
        }
    }
}
